from rest_framework import serializers

from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.utils import is_non_string_iterable


class MaybeListCharField(serializers.CharField):
    def to_internal_value(self, data):
        if is_non_string_iterable(data):
            return data
        return super(MaybeListCharField, self).to_internal_value(data)


class SecurityReferencValidator(object):
    context = None

    def __call__(self, value):
        if self.context is None:
            raise ValueError("Missing validation context")
        if value not in self.context.get('securityDefinitions', {}):
            raise ValidationError(
                MESSAGES['unknown_reference']['security'].format(value),
            )

    def set_context(self, parent):
        self.context = parent.context


class SecurityRequirementReferenceField(serializers.CharField):
    """
    Field that references a defined security scheme declared in the Security
    Definitions.
    """
    def __init__(self, *args, **kwargs):
        super(SecurityRequirementReferenceField, self).__init__(*args, **kwargs)
        self.validators.append(SecurityReferencValidator())


class DefaultField(serializers.Field):
    def to_internal_value(self, data):
        return data



from __future__ import unicode_literals

import functools
import six

from rest_framework import serializers

from drf_compound_fields.fields import ListField

from flex.exceptions import (
    ValidationError,
    ErrorDict,
)
from flex.context_managers import ErrorCollection
from flex.serializers.validators import (
    host_validator,
    path_validator,
    scheme_validator,
    mimetype_validator,
    string_type_validator,
)
from flex.constants import (
    PATH,
    REQUEST_METHODS,
)
from flex.validation.common import (
    validate_object,
)
from flex.validation.schema import (
    construct_schema_validators,
)
from flex.paths import (
    get_parameter_names_from_path,
)
from flex.parameters import (
    filter_parameters,
    merge_parameter_lists,
    dereference_parameter_list,
)
from flex.error_messages import MESSAGES

from .fields import (
    SecurityRequirementReferenceField,
)
from .common import (
    HomogenousDictSerializer,
    BaseResponseSerializer,
    BaseParameterSerializer,
    BaseSchemaSerializer,
    BaseItemsSerializer,
    BaseHeaderSerializer,
)


class InfoSerializer(serializers.Serializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#infoObject
    """
    title = serializers.CharField()
    description = serializers.CharField(required=False)
    termsOfService = serializers.CharField(required=False)
    contact = serializers.CharField(required=False)
    license = serializers.CharField(required=False)
    version = serializers.CharField(required=False)


class ItemsSerializer(BaseItemsSerializer):
    default_error_messages = {
        'unknown_reference': 'Unknown definition reference `{0}`',
    }

    def from_native(self, data, files):
        if isinstance(data, six.string_types):
            definitions = self.context.get('definitions', {})
            if data not in definitions:
                raise ValidationError(
                    MESSAGES['unknown_reference']['definition'].format(data),
                )
            return data
        return super(ItemsSerializer, self).from_native(data, files)


class HeaderSerializer(BaseHeaderSerializer):
    items = ItemsSerializer(required=False, many=True)


class HeadersSerializer(HomogenousDictSerializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#headersObject
    """
    value_serializer_class = HeaderSerializer


class SchemaSerializer(BaseSchemaSerializer):
    default_error_messages = {
        'unknown_reference': 'Unknown definition reference `{0}`'
    }

    def validate(self, attrs):
        errors = ErrorDict()

        if '$ref' in attrs:
            definitions = self.context.get('definitions', {})
            if attrs['$ref'] not in definitions:
                errors['$ref'].add_error(
                    self.error_messages['unknown_reference'].format(attrs['$ref']),
                )

        if errors:
            raise ValidationError(errors)
        return super(SchemaSerializer, self).validate(attrs)

    def save_object(self, obj, **kwargs):
        validators = construct_schema_validators(obj, self.context)
        self.object = functools.partial(validate_object, validators=validators)


class ResponseSerializer(BaseResponseSerializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#responseObject
    """
    schema = SchemaSerializer(required=False)
    headers = HeadersSerializer(required=False)
    # TODO: how do we do examples
    # examples =


class ResponsesSerializer(HomogenousDictSerializer):
    value_serializer_class = ResponseSerializer


class SecuritySerializer(HomogenousDictSerializer):
    value_serializer_class = SecurityRequirementReferenceField


class ParameterSerializer(BaseParameterSerializer):
    schema = SchemaSerializer(required=False)
    items = ItemsSerializer(required=False, many=True)

    @property
    def many(self):
        return True

    @many.setter
    def many(self, value):
        pass

    def from_native(self, data, files=None):
        if isinstance(data, six.string_types):
            try:
                self.validate_reference(data)
            except ValidationError as err:
                assert not self._errors
                self._errors = err.messages
                return
            else:
                return data
        return super(ParameterSerializer, self).from_native(data, files)

    def validate_reference(self, reference):
        if reference not in self.context.get('parameters', {}):
            raise ValidationError(
                MESSAGES['unknown_reference']['parameter'].format(reference),
            )

    def save(self):
        return self.object


class OperationSerializer(serializers.Serializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#operationObject
    """
    tags = ListField(required=False, validators=[string_type_validator])
    summary = serializers.CharField(required=False)
    description = serializers.CharField(required=False)
    externalDocs = serializers.CharField(required=False)
    operationId = serializers.CharField(required=False)
    consumes = ListField(required=False, validators=[mimetype_validator])
    produces = ListField(required=False, validators=[mimetype_validator])
    parameters = ParameterSerializer(required=False, many=True)
    responses = ResponsesSerializer()
    schemes = ListField(required=False, validators=[scheme_validator])
    deprecated = serializers.BooleanField(required=False)
    security = SecuritySerializer(required=False)


class PathItemSerializer(serializers.Serializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#pathsObject
    """
    # TODO. reference path item objects from definitions.
    # TODO. how is this supposted to work.  The swagger spec doesn't account
    # for a definitions location for PathItem definitions?
    # _ref = serializers.CharField(source='$ref')
    get = OperationSerializer(required=False)
    put = OperationSerializer(required=False)
    post = OperationSerializer(required=False)
    delete = OperationSerializer(required=False)
    options = OperationSerializer(required=False)
    head = OperationSerializer(required=False)
    patch = OperationSerializer(required=False)
    # TODO: these can be a parameters reference object.
    parameters = ParameterSerializer(required=False, many=True)


class TagSerializer(serializers.Serializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#tagObject
    """
    name = serializers.CharField()
    description = serializers.CharField(required=False)
    externalDocs = serializers.CharField(required=False)


class PropertiesSerializer(HomogenousDictSerializer):
    value_serializer_class = SchemaSerializer


# These fields include recursive use of the `SchemaSerializer` so they have to
# be attached after the `SchemaSerializer` class has been created.
SchemaSerializer.base_fields['properties'] = PropertiesSerializer(required=False)
SchemaSerializer.base_fields['items'] = ItemsSerializer(required=False, many=True)
SchemaSerializer.base_fields['allOf'] = SchemaSerializer(required=False, many=True)


class PathsSerializer(HomogenousDictSerializer):
    value_serializer_class = PathItemSerializer
    allow_empty = True

    def validate(self, attrs):
        with ErrorCollection() as errors:
            for api_path, path_definition in attrs.items():
                path_parameter_names = set(get_parameter_names_from_path(api_path))

                if path_definition is None:
                    continue

                api_path_level_parameters = dereference_parameter_list(
                    path_definition.get('parameters', []),
                    parameter_definitions=self.context.get('parameters', {}),
                )

                path_request_methods = set(REQUEST_METHODS).intersection(
                    path_definition.keys(),
                )

                if not path_request_methods:
                    for parameter in api_path_level_parameters:
                        if parameter['name'] not in path_parameter_names:
                            errors[api_path].add_error(
                                MESSAGES["path"]["missing_parameter"].format(
                                    parameter['name'], api_path,
                                ),
                            )

                for method, operation_definition in path_definition.items():
                    if method not in REQUEST_METHODS:
                        continue
                    if operation_definition is None:
                        operation_definition = {}
                    operation_level_parameters = dereference_parameter_list(
                        operation_definition.get('parameters', []),
                        parameter_definitions=self.context.get('parameters', {}),
                    )
                    parameters_in_path = filter_parameters(
                        merge_parameter_lists(
                            api_path_level_parameters,
                            operation_level_parameters,
                        ),
                        in_=PATH,
                    )

                    for parameter in parameters_in_path:
                        if parameter['name'] not in path_parameter_names:
                            key = "{method}:{api_path}".format(
                                method=method.upper(),
                                api_path=api_path,
                            )
                            errors[key].add_error(
                                MESSAGES["path"]["missing_parameter"].format(
                                    parameter['name'], api_path,
                                ),
                            )

        return super(PathsSerializer, self).validate(attrs)

    def save_object(self, obj, **kwargs):
        self.object = obj


class SwaggerSerializer(serializers.Serializer):
    """
    Primary Serializer for swagger schema
    """
    swagger = serializers.ChoiceField(
        choices=[
            ('2.0', '2.0'),
        ],
    )
    info = InfoSerializer()
    host = serializers.CharField(
        required=False,
        validators=[host_validator],
    )
    basePath = serializers.CharField(
        required=False,
        validators=[path_validator],
    )
    schemes = ListField(required=False, validators=[scheme_validator])
    consumes = ListField(required=False, validators=[mimetype_validator])
    produces = ListField(required=False, validators=[mimetype_validator])

    paths = PathsSerializer()

    security = SecuritySerializer(required=False)

    tags = TagSerializer(required=False, many=True)
    externalDocs = serializers.CharField(required=False)

    def save_object(self, obj, **kwargs):
        self.object = obj



import rest_framework

__all__ = [
    'SwaggerSerializer',
    'SchemaSerializer',
    'PathsSerializer',
    'ParameterSerializer',
    'HeaderSerializer',
]

if rest_framework.__version__ >= '3.0.0':
    from .drf_3.core import (
        SwaggerSerializer,
        SchemaSerializer,
        PathsSerializer,
        ParameterSerializer,
        HeaderSerializer,
    )
else:
    from .drf_2.core import (
        SwaggerSerializer,
        SchemaSerializer,
        PathsSerializer,
        ParameterSerializer,
        HeaderSerializer,
    )



import functools

from flex.validation.parameter import (
    validate_path_parameters,
)


def generate_path_parameters_validator(api_path, path_parameters, context):
    """
    Generates a validator function that given a path, validates that it against
    the path parameters
    """
    path_parameter_validator = functools.partial(
        validate_path_parameters,
        api_path=api_path,
        path_parameters=path_parameters,
        context=context,
    )
    return path_parameter_validator



from flex.exceptions import ValidationError
from flex.context_managers import ErrorCollection
from flex.validation.operation import (
    construct_operation_validators,
    validate_operation,
)
from flex.validation.common import (
    validate_request_method_to_operation,
    validate_path_to_api_path,
)


def validate_request(request, schema):
    """
    Request validation does the following steps.

       1. validate that the path matches one of the defined paths in the schema.
       2. validate that the request method conforms to a supported methods for the given path.
       3. validate that the request parameters conform to the parameter
          definitions for the operation definition.
    """
    with ErrorCollection() as errors:
        # 1
        try:
            api_path = validate_path_to_api_path(
                path=request.path,
                **schema
            )
        except ValidationError as err:
            errors['path'].add_error(err.detail)
            return  # this causes an exception to be raised since errors is no longer falsy.

        path_definition = schema['paths'][api_path] or {}

        if not path_definition:
            # TODO: is it valid to not have a definition for a path?
            return

        # 2
        try:
            operation_definition = validate_request_method_to_operation(
                request_method=request.method,
                path_definition=path_definition,
            )
        except ValidationError as err:
            errors['method'].add_error(err.detail)
            return

        if operation_definition is None:
            # TODO: is this compliant with swagger, can path operations have a null
            # definition?
            return

        # 3
        operation_validators = construct_operation_validators(
            api_path=api_path,
            path_definition=path_definition,
            operation_definition=operation_definition,
            context=schema,
        )
        try:
            validate_operation(request, operation_validators)
        except ValidationError as err:
            errors['method'].add_error(err.detail)



import rest_framework

__all__ = [
    'TypedDefaultMixin',
    'TranslateValidationErrorMixin',
]

if rest_framework.__version__ >= '3.0.0':
    from .drf_3.mixins import (
        TypedDefaultMixin,
        TranslateValidationErrorMixin,
    )
else:
    from .drf_2.mixins import (
        TypedDefaultMixin,
        TranslateValidationErrorMixin,
    )



import six


class IntKeyedDict(dict):
    """
    Returns a dictionary that treats integers and strings as the same key for
    retrieval.  Useful for the return value from the `ResponsesSerializer` whos
    keys are the response status codes.
    """
    def __getitem__(self, key):
        try:
            return super(IntKeyedDict, self).__getitem__(key)
        except KeyError:
            if isinstance(key, int):
                key = str(key)
            elif isinstance(key, six.string_types) and key.isdigit():
                key = int(key)
            else:
                raise

            return super(IntKeyedDict, self).__getitem__(key)



# RFC 2822 - style email validation for Python
# (c) 2012 Syrus Akbary <me@syrusakbary.com>
# Extended from (c) 2011 Noel Bush <noel@aitools.org>
# for support of mx and user check
# This code is made available to you under the GNU LGPL v3.
#
# This module provides a single method, valid_email_address(),
# which returns True or False to indicate whether a given address
# is valid according to the 'addr-spec' part of the specification
# given in RFC 2822.  Ideally, we would like to find this
# in some other library, already thoroughly tested and well-
# maintained.  The standard Python library email.utils
# contains a parse_addr() function, but it is not sufficient
# to detect many malformed addresses.
#
# This implementation aims to be faithful to the RFC, with the
# exception of a circular definition (see comments below), and
# with the omission of the pattern components marked as "obsolete".

import re
import smtplib
import logging
import socket

try:
    raw_input
except NameError:
    def raw_input(prompt=''):
        return input(prompt)

try:
    import DNS
    ServerError = DNS.ServerError
    DNS.DiscoverNameServers()
except ImportError:
    DNS = None

    class ServerError(Exception):
        pass

# All we are really doing is comparing the input string to one
# gigantic regular expression.  But building that regexp, and
# ensuring its correctness, is made much easier by assembling it
# from the "tokens" defined by the RFC.  Each of these tokens is
# tested in the accompanying unit test file.
#
# The section of RFC 2822 from which each pattern component is
# derived is given in an accompanying comment.
#
# (To make things simple, every string below is given as 'raw',
# even when it's not strictly necessary.  This way we don't forget
# when it is necessary.)
#
WSP = r'[ \t]'                                       # see 2.2.2. Structured Header Field Bodies
CRLF = r'(?:\r\n)'                                   # see 2.2.3. Long Header Fields
NO_WS_CTL = r'\x01-\x08\x0b\x0c\x0f-\x1f\x7f'        # see 3.2.1. Primitive Tokens
QUOTED_PAIR = r'(?:\\.)'                             # see 3.2.2. Quoted characters
FWS = r'(?:(?:' + WSP + r'*' + CRLF + r')?' + \
      WSP + r'+)'                                    # see 3.2.3. Folding white space and comments
CTEXT = r'[' + NO_WS_CTL + \
        r'\x21-\x27\x2a-\x5b\x5d-\x7e]'              # see 3.2.3
CCONTENT = r'(?:' + CTEXT + r'|' + \
           QUOTED_PAIR + r')'                        # see 3.2.3 (NB: The RFC includes COMMENT here
# as well, but that would be circular.)
COMMENT = r'\((?:' + FWS + r'?' + CCONTENT + \
          r')*' + FWS + r'?\)'                       # see 3.2.3
CFWS = r'(?:' + FWS + r'?' + COMMENT + ')*(?:' + \
       FWS + '?' + COMMENT + '|' + FWS + ')'         # see 3.2.3
ATEXT = r'[\w!#$%&\'\*\+\-/=\?\^`\{\|\}~]'           # see 3.2.4. Atom
ATOM = CFWS + r'?' + ATEXT + r'+' + CFWS + r'?'      # see 3.2.4
DOT_ATOM_TEXT = ATEXT + r'+(?:\.' + ATEXT + r'+)*'   # see 3.2.4
DOT_ATOM = CFWS + r'?' + DOT_ATOM_TEXT + CFWS + r'?' # see 3.2.4
QTEXT = r'[' + NO_WS_CTL + \
        r'\x21\x23-\x5b\x5d-\x7e]'                   # see 3.2.5. Quoted strings
QCONTENT = r'(?:' + QTEXT + r'|' + \
           QUOTED_PAIR + r')'                        # see 3.2.5
QUOTED_STRING = CFWS + r'?' + r'"(?:' + FWS + \
                r'?' + QCONTENT + r')*' + FWS + \
                r'?' + r'"' + CFWS + r'?'
LOCAL_PART = r'(?:' + DOT_ATOM + r'|' + \
             QUOTED_STRING + r')'                    # see 3.4.1. Addr-spec specification
DTEXT = r'[' + NO_WS_CTL + r'\x21-\x5a\x5e-\x7e]'    # see 3.4.1
DCONTENT = r'(?:' + DTEXT + r'|' + \
           QUOTED_PAIR + r')'                        # see 3.4.1
DOMAIN_LITERAL = CFWS + r'?' + r'\[' + \
                 r'(?:' + FWS + r'?' + DCONTENT + \
                 r')*' + FWS + r'?\]' + CFWS + r'?'  # see 3.4.1
DOMAIN = r'(?:' + DOT_ATOM + r'|' + \
         DOMAIN_LITERAL + r')'                       # see 3.4.1
ADDR_SPEC = LOCAL_PART + r'@' + DOMAIN               # see 3.4.1

# A valid address will match exactly the 3.4.1 addr-spec.
VALID_ADDRESS_REGEXP = '^' + ADDR_SPEC + '$'

MX_DNS_CACHE = {}
MX_CHECK_CACHE = {}


def get_mx_ip(hostname):
    if hostname not in MX_DNS_CACHE:
        try:
            MX_DNS_CACHE[hostname] = DNS.mxlookup(hostname)
        except ServerError as e:
            if e.rcode == 3:  # NXDOMAIN (Non-Existent Domain)
                MX_DNS_CACHE[hostname] = None
            else:
                raise

    return MX_DNS_CACHE[hostname]


def validate_email(email, check_mx=False, verify=False, debug=False, smtp_timeout=10):
    """Indicate whether the given string is a valid email address
    according to the 'addr-spec' portion of RFC 2822 (see section
    3.4.1).  Parts of the spec that are marked obsolete are *not*
    included in this test, and certain arcane constructions that
    depend on circular definitions in the spec may not pass, but in
    general this should correctly identify any email address likely
    to be in use as of 2011."""
    if debug:
        logger = logging.getLogger('validate_email')
        logger.setLevel(logging.DEBUG)
    else:
        logger = None

    try:
        assert re.match(VALID_ADDRESS_REGEXP, email) is not None
        check_mx |= verify
        if check_mx:
            if not DNS:
                raise Exception('For check the mx records or check if the email exists you must '
                                'have installed pyDNS python package')
            hostname = email[email.find('@') + 1:]
            mx_hosts = get_mx_ip(hostname)
            if mx_hosts is None:
                return False
            for mx in mx_hosts:
                try:
                    if not verify and mx[1] in MX_CHECK_CACHE:
                        return MX_CHECK_CACHE[mx[1]]
                    smtp = smtplib.SMTP(timeout=smtp_timeout)
                    smtp.connect(mx[1])
                    MX_CHECK_CACHE[mx[1]] = True
                    if not verify:
                        try:
                            smtp.quit()
                        except smtplib.SMTPServerDisconnected:
                            pass
                        return True
                    status, _ = smtp.helo()
                    if status != 250:
                        smtp.quit()
                        if debug:
                            logger.debug(u'%s answer: %s - %s', mx[1], status, _)
                        continue
                    smtp.mail('')
                    status, _ = smtp.rcpt(email)
                    if status == 250:
                        smtp.quit()
                        return True
                    if debug:
                        logger.debug(u'%s answer: %s - %s', mx[1], status, _)
                    smtp.quit()
                except smtplib.SMTPServerDisconnected:  # Server not permits verify user
                    if debug:
                        logger.debug(u'%s disconected.', mx[1])
                except smtplib.SMTPConnectError:
                    if debug:
                        logger.debug(u'Unable to connect to %s.', mx[1])
            return None
    except AssertionError:
        return False
    except (ServerError, socket.error) as e:
        if debug:
            logger.debug('ServerError or socket.error exception raised (%s).', e)
        return None
    return True

if __name__ == "__main__":
    import time
    while True:
        email = raw_input('Enter email for validation: ')

        mx = raw_input('Validate MX record? [yN] ')
        if mx.strip().lower() == 'y':
            mx = True
        else:
            mx = False

        validate = raw_input('Try to contact server for address validation? [yN] ')
        if validate.strip().lower() == 'y':
            validate = True
        else:
            validate = False

        logging.basicConfig()

        result = validate_email(email, mx, validate, debug=True, smtp_timeout=1)
        if result:
            print("Valid!")
        elif result is None:
            print("I'm not sure.")
        else:
            print("Invalid!")

        time.sleep(1)


# import sys

# sys.modules[__name__],sys.modules['validate_email_module'] = validate_email,sys.modules[__name__]
# from validate_email_module import *



import functools
import operator

from flex.exceptions import ValidationError
from flex.utils import chain_reduce_partial
from flex.context_managers import ErrorCollection
from flex.http import (
    Request,
)
from flex.constants import (
    QUERY,
    PATH,
    EMPTY,
    HEADER,
)
from flex.parameters import (
    filter_parameters,
    merge_parameter_lists,
    dereference_parameter_list,
)
from flex.validation.parameter import (
    validate_query_parameters,
)
from flex.validation.header import (
    construct_header_validators,
)
from flex.validation.path import (
    generate_path_parameters_validator,
)
from flex.validation.common import (
    validate_object,
    generate_value_processor,
)


def validate_operation(request, validators):
    with ErrorCollection() as errors:
        for key, validator in validators.items():
            try:
                validator(request)
            except ValidationError as err:
                errors[key].add_error(err.detail)


def validate_request_content_type(request, content_types):
    assert isinstance(request, Request)
    # TODO: is it correct to skip validation for a null content_type?
    if request.content_type and request.content_type not in content_types:
        raise ValidationError(
            'Invalid content type `{0}`.  Must be one of `{1}`.'.format(
                request.content_type, content_types,
            ),
        )


def generate_request_content_type_validator(consumes, **kwargs):
    validator = functools.partial(
        validate_request_content_type,
        content_types=consumes,
    )
    return validator


def generate_query_parameters_validator(query_parameters, context):
    query_parameter_validator = functools.partial(
        validate_query_parameters,
        query_parameters=query_parameters,
        context=context,
    )
    return chain_reduce_partial(
        operator.attrgetter('query_data'),
        query_parameter_validator,
    )


def generate_header_validator(headers, context, **kwargs):
    """
    Generates a validation function that will validate a dictionary of headers.
    """
    validators = {}
    for header_definition in headers:
        header_processor = generate_value_processor(
            context=context,
            **header_definition
        )
        header_validator = functools.partial(
            validate_object,
            validators=construct_header_validators(header_definition, context=context),
        )
        validators[header_definition['name']] = chain_reduce_partial(
            operator.methodcaller('get', header_definition['name'], EMPTY),
            header_processor,
            header_validator,
        )
    return functools.partial(validate_object, validators=validators)


def generate_parameters_validator(api_path, path_definition, parameters,
                                  context, **kwargs):
    """
    Generates a validator function to validate.

    - request.path against the path parameters.
    - request.query against the query parameters.
    - request.headers against the header parameters.
    - TODO: request.body against the body parameters.
    - TODO: request.formData against any form data.
    """
    # TODO: figure out how to merge this with the same code in response
    # validation.
    validators = {}
    parameter_definitions = context.get('parameters', {})
    path_level_parameters = dereference_parameter_list(
        path_definition.get('parameters', []),
        parameter_definitions,
    )
    operation_level_parameters = dereference_parameter_list(
        parameters,
        parameter_definitions,
    )

    all_parameters = merge_parameter_lists(
        path_level_parameters,
        operation_level_parameters,
    )

    # PATH
    in_path_parameters = filter_parameters(all_parameters, in_=PATH)
    validators['path'] = chain_reduce_partial(
        operator.attrgetter('path'),
        generate_path_parameters_validator(api_path, in_path_parameters, context),
    )

    # QUERY
    in_query_parameters = filter_parameters(all_parameters, in_=QUERY)
    validators['query'] = generate_query_parameters_validator(in_query_parameters, context)

    # HEADERS
    in_header_parameters = filter_parameters(all_parameters, in_=HEADER)
    validators['headers'] = chain_reduce_partial(
        operator.attrgetter('headers'),
        generate_header_validator(in_header_parameters, context),
    )

    return functools.partial(validate_object, validators=validators)


validator_mapping = {
    'consumes': generate_request_content_type_validator,
    'parameters': generate_parameters_validator,
    'headers': generate_header_validator,
}


def construct_operation_validators(api_path, path_definition, operation_definition, context):
    """
    - consumes (did the request conform to the content types this api consumes)
    - produces (did the response conform to the content types this endpoint produces)
    - parameters (did the parameters of this request validate)
      TODO: move path parameter validation to here, because each operation
            can override any of the path level parameters.
    - schemes (was the request scheme correct)
    - security: TODO since security isn't yet implemented.
    """
    validators = {}

    # sanity check
    assert 'context' not in operation_definition
    assert 'api_path' not in operation_definition
    assert 'path_definition' not in operation_definition

    for key in operation_definition.keys():
        if key not in validator_mapping:
            # TODO: is this the right thing to do?
            continue
        validators[key] = validator_mapping[key](
            context=context,
            api_path=api_path,
            path_definition=path_definition,
            **operation_definition
        )

    # Global defaults
    if 'consumes' in context and 'consumes' not in validators:
        validators['consumes'] = validator_mapping['consumes'](**context)
    if 'parameters' in path_definition and 'parameters' not in validators:
        validators['parameters'] = validator_mapping['parameters'](
            context=context,
            api_path=api_path,
            path_definition=path_definition,
            parameters=path_definition['parameters'],
            **operation_definition
        )

    return validators



import functools
import math
import collections
import numbers

import six

from flex.constants import (
    PRIMATIVE_TYPES,
    NULL,
    BOOLEAN,
    INTEGER,
    NUMBER,
    STRING,
    ARRAY,
    OBJECT,
    TRUE_VALUES,
    FALSE_VALUES,
)


def is_non_string_iterable(value):
    if six.PY2:
        string_types = six.string_types
    else:
        string_types = (six.binary_type, six.text_type)
    return not isinstance(value, string_types) and hasattr(value, '__iter__')


def is_value_of_type(value, type_):
    if type_ not in PRIMATIVE_TYPES:
        raise ValueError("Unknown type: {0}".format(type_))

    if type_ == ARRAY and is_value_of_type(value, STRING):
        return False

    if type_ in (INTEGER, NUMBER) and is_value_of_type(value, BOOLEAN):
        return False

    return isinstance(value, PRIMATIVE_TYPES[type_])


def is_value_of_any_type(value, types):
    return any(is_value_of_type(value, type_) for type_ in types)


def cast_value_to_type(value, type_):
    if type_ == STRING:
        return six.text_type(value)
    elif type_ == INTEGER:
        return int(value)
    elif type_ == NUMBER:
        return float(value)
    elif type_ == ARRAY:
        return list(value)
    elif type_ == OBJECT:
        return dict(value)
    elif type_ == BOOLEAN:
        if value in TRUE_VALUES:
            return True
        elif value in FALSE_VALUES:
            return False
        else:
            raise TypeError("Invalid value for boolean: `{0}`".format(repr(value)))
    # TODO: the only thing left is null type.
    return PRIMATIVE_TYPES[type_][0](value)


def get_type_for_value(value):
    if value is None:
        return NULL
    if isinstance(value, PRIMATIVE_TYPES[BOOLEAN]):
        return BOOLEAN
    elif isinstance(value, PRIMATIVE_TYPES[INTEGER]):
        return INTEGER
    elif isinstance(value, PRIMATIVE_TYPES[NUMBER]):
        return NUMBER
    elif isinstance(value, PRIMATIVE_TYPES[STRING]):
        return STRING
    elif isinstance(value, PRIMATIVE_TYPES[ARRAY]):
        return ARRAY
    elif isinstance(value, PRIMATIVE_TYPES[OBJECT]):
        return OBJECT
    else:
        raise ValueError("Unable to identify type of {0}".format(repr(value)))


def is_single_item_iterable(value):
    if is_non_string_iterable(value):
        if isinstance(value, collections.Sequence):
            if len(value) == 1:
                return True
    return False


def indent_message(message, indent, prefix='', suffix=''):
    return "{indent}{prefix}{message}{suffix}".format(
        indent=' ' * indent,
        prefix=prefix,
        message=message,
        suffix=suffix,
    )


SINGULAR_TYPES = six.string_types + (numbers.Number,)


def format_errors(errors, indent=0, prefix='', suffix=''):
    """
    string: "example"

        "example"

    dict:
        "example":
            -

    """
    if is_single_item_iterable(errors):
        errors = errors[0]
    if isinstance(errors, SINGULAR_TYPES):
        yield indent_message(repr(errors), indent, prefix=prefix, suffix=suffix)

    elif isinstance(errors, collections.Mapping):
        for key, value in errors.items():
            assert isinstance(key, SINGULAR_TYPES), type(key)
            if isinstance(value, SINGULAR_TYPES):
                message = "{0}: {1}".format(repr(key), repr(value))
                yield indent_message(message, indent, prefix=prefix, suffix=suffix)
            else:
                yield indent_message(repr(key), indent, prefix=prefix, suffix=':')
                for message in format_errors(value, indent + 4, prefix='- '):
                    yield message

    elif is_non_string_iterable(errors):
        # for making the rhs of the numbers line up
        extra_indent = int(math.ceil(math.log10(len(errors)))) + 2
        for index, value in enumerate(errors):
            list_prefix = "{0}. ".format(index)
            messages = format_errors(
                value,
                indent=indent + extra_indent - len(list_prefix),
                prefix=list_prefix,
            )
            for message in messages:
                yield message
    else:
        assert False, "should not be possible"


def prettify_errors(errors):
    return '\n'.join(format_errors(errors))


def chain_reduce_partial(*functions):
    """
    Given an iterable of functions, returns a callable that takes a value and
    passes it through all of the given functions in order.

    def a(x):
        ...

    def b(x):
        ...

    c = chain_reduce_partial(a, b)

    This is equivilent to

    def c(x):
        return b(a(x))
    """
    return functools.partial(
        functools.reduce,
        lambda value, fn: fn(value),
        functions,
    )



from flex.exceptions import (
    ValidationError,
    ErrorDict,
)


class ErrorCollection(object):
    def __init__(self, message='Invalid'):
        self.message = message
        self.errors = ErrorDict()

    def __enter__(self):
        return self.errors

    def __exit__(self, type_, value, traceback):
        if any((type_, value, traceback)):
            if issubclass(type_, ValidationError):
                self.errors = value.detail
            else:
                return False
        if self.errors:
            raise ValidationError(dict(self.errors))



from flex.validation.common import (
    generate_type_validator,
    generate_format_validator,
    generate_maximum_validator,
    generate_minimum_validator,
    generate_min_length_validator,
    generate_max_length_validator,
    generate_pattern_validator,
    generate_min_items_validator,
    generate_max_items_validator,
    generate_unique_items_validator,
    generate_enum_validator,
    generate_multiple_of_validator,
)
from flex.validation.schema import (
    generate_items_validator,
)


validator_mapping = {
    # TODO: for array types, the value will come through as a string that *may*
    # be delimited.  Need to split on the delimiter and turn it into an array
    # before running validation.
    'type': generate_type_validator,
    'format': generate_format_validator,
    'items': generate_items_validator,
    'maximum': generate_maximum_validator,
    'minimum': generate_minimum_validator,
    'minLength': generate_min_length_validator,
    'maxLength': generate_max_length_validator,
    'pattern': generate_pattern_validator,
    'minItems': generate_min_items_validator,
    'maxItems': generate_max_items_validator,
    'uniqueItems': generate_unique_items_validator,
    'enum': generate_enum_validator,
    'multipleOf': generate_multiple_of_validator,
    # TODO:
    # collectionFormat
}


def construct_header_validators(header_definition, context):
    validators = {}

    for key in header_definition:
        if key in validator_mapping:
            validators[key] = validator_mapping[key](context=context, **header_definition)

    return validators



import itertools
import collections
import functools
import operator

import six

from flex.exceptions import (
    ValidationError,
    ErrorList,
)
from flex.error_messages import MESSAGES
from flex.constants import (
    OBJECT,
    EMPTY,
)
from flex.decorators import skip_if_not_of_type
from flex.validation.common import (
    skip_if_empty,
    generate_type_validator,
    generate_format_validator,
    generate_required_validator,
    generate_multiple_of_validator,
    generate_minimum_validator,
    generate_maximum_validator,
    generate_min_length_validator,
    generate_max_length_validator,
    generate_min_items_validator,
    generate_max_items_validator,
    generate_unique_items_validator,
    generate_pattern_validator,
    generate_enum_validator,
    validate_object,
)
from flex.utils import (
    chain_reduce_partial,
)


@skip_if_empty
@skip_if_not_of_type(OBJECT)
def validate_min_properties(value, minimum):
    if len(value.keys()) < minimum:
        raise ValidationError(
            MESSAGES['min_properties']['invalid'].format(
                minimum, len(value.keys()),
            ),
        )


def generate_min_properties_validator(minProperties, **kwargs):
    return functools.partial(validate_min_properties, minimum=minProperties)


@skip_if_empty
@skip_if_not_of_type(OBJECT)
def validate_max_properties(value, maximum):
    if len(value.keys()) > maximum:
        raise ValidationError(
            MESSAGES['max_properties']['invalid'].format(
                maximum, len(value.keys()),
            ),
        )


def generate_max_properties_validator(maxProperties, **kwargs):
    return functools.partial(validate_max_properties, maximum=maxProperties)


def construct_items_validators(items, context):
    if isinstance(items, collections.Mapping):
        items_validators = construct_schema_validators(
            items,
            context,
        )
    elif isinstance(items, six.string_types):
        items_validators = {
            '$ref': LazyReferenceValidator(items, context),
        }
    else:
        assert 'Should not be possible'
    return items_validators


def validate_items(objs, validators):
    errors = ErrorList()
    for obj, validator in zip(objs, validators):
        try:
            validate_object(obj, validator)
        except ValidationError as e:
            errors.add_error(e.detail)

    if errors:
        raise ValidationError(errors)


def generate_items_validator(items, context, **kwargs):
    if isinstance(items, collections.Mapping) or isinstance(items, six.string_types):
        # If items is a reference or a schema, we pass it through as an
        # ever repeating list of the same validation dictionary, thus
        # validating all of the objects against the same schema.
        items_validators = itertools.repeat(construct_items_validators(
            items,
            context,
        ))
    elif isinstance(items, collections.Sequence):
        # We generate a list of validator dictionaries and then chain it
        # with an empty schema that repeats forever.  This ensures that if
        # the array of objects to be validated is longer than the array of
        # validators, then the extra elements will always validate since
        # they will be validated against an empty schema.
        items_validators = itertools.chain(
            map(functools.partial(construct_items_validators, context=context), items),
            itertools.repeat({}),
        )
    else:
        assert "Should not be possible"
    return functools.partial(
        validate_items, validators=items_validators,
    )


validator_mapping = {
    'type': generate_type_validator,
    'multipleOf': generate_multiple_of_validator,
    'minimum': generate_minimum_validator,
    'maximum': generate_maximum_validator,
    'minLength': generate_min_length_validator,
    'maxLength': generate_max_length_validator,
    'minItems': generate_min_items_validator,
    'maxItems': generate_max_items_validator,
    'uniqueItems': generate_unique_items_validator,
    'enum': generate_enum_validator,
    'minProperties': generate_min_properties_validator,
    'maxProperties': generate_max_properties_validator,
    'pattern': generate_pattern_validator,
    'format': generate_format_validator,
    'required': generate_required_validator,
    'items': generate_items_validator,
}


class LazyReferenceValidator(object):
    """
    This class acts as a lazy validator for references in schemas to prevent an
    infinite recursion error when a schema references itself, or there is a
    reference loop between more than one schema.

    The validator is only constructed if validator is needed.
    """
    def __init__(self, reference, context):
        # TODO: something better than this assertion
        assert 'definitions' in context
        assert reference in context['definitions']
        self.reference = reference
        self.context = context

    def __call__(self, value):
        return validate_object(value, self.validators)

    @property
    def validators(self):
        return construct_schema_validators(
            self.context['definitions'][self.reference],
            self.context,
        )

    def items(self):
        return self.validators.items()


def construct_schema_validators(schema, context):
    """
    Given a schema object, construct a dictionary of validators needed to
    validate a response matching the given schema.

    Special Cases:
        - $ref:
            These validators need to be Lazily evaluating so that circular
            validation dependencies do not result in an infinitely deep
            validation chain.
        - properties:
            These validators are meant to apply to properties of the object
            being validated rather than the object itself.  In this case, we
            need recurse back into this function to generate a dictionary of
            validators for the property.
    """
    validators = {}
    if '$ref' in schema:
        validators['$ref'] = LazyReferenceValidator(
            schema['$ref'],
            context,
        )
    if 'properties' in schema:
        # I don't know why this set intersection is enforced...?  Why did I do this.
        intersection = set(schema['properties'].keys()).intersection(schema.keys())
        assert not intersection

        for property_, property_schema in schema['properties'].items():
            property_validators = construct_schema_validators(
                property_schema,
                context,
            )
            validators[property_] = skip_if_empty(skip_if_not_of_type(OBJECT)(
                chain_reduce_partial(
                    operator.methodcaller('get', property_, EMPTY),
                    functools.partial(validate_object, validators=property_validators),
                ),
            ))
    assert 'context' not in schema
    for key in schema:
        if key in validator_mapping:
            validators[key] = validator_mapping[key](context=context, **schema)
    return validators



import copy
import six

from django.core.validators import (
    MinValueValidator,
)

from rest_framework import serializers
from rest_framework.utils import representation

from flex.error_messages import MESSAGES
from flex.exceptions import (
    ValidationError,
    ErrorDict,
)
from flex.utils import (
    is_value_of_any_type,
)
from flex.serializers.validators import (
    type_validator,
    format_validator,
    parameter_in_validator,
    collection_format_validator,
    regex_validator,
    is_array_validator,
    header_type_validator,
)
from flex.constants import (
    BODY,
    PATH,
    CSV,
    QUERY,
    FORM_DATA,
    MULTI,
    ARRAY,
    INTEGER,
    NUMBER,
    STRING,
    OBJECT,
)

from .fields import (
    MaybeListCharField,
    DefaultField,
)
from .mixins import (
    TypedDefaultMixin,
)


def make_serializer_repr_safe(serializer):
    """
    Since some of the serializers have recursive fields on them
    (`SchemaSerializer`), we need to avoid infinite recursion while
    trying to repr them.
    """
    serializer_copy = copy.deepcopy(serializer)

    for field_name in serializer_copy.fields.keys():
        field = serializer_copy.fields[field_name]
        if hasattr(field, 'child'):
            field = field.child
        if getattr(field, '_is_unsafe_to_repr', False):
            serializer_copy.fields.pop(field_name)
    return serializer_copy


class ReprSafeListSerializer(serializers.ListSerializer):
    def __repr__(self):
        return representation.serializer_repr(
            make_serializer_repr_safe(self.child),
            indent=1,
        )


class ReprSafeSerializer(object):
    def __repr__(self):
        """
        Since some of the serializers have recursive fields on them
        (`SchemaSerializer`), we need to avoid infinite recursion while
        trying to repr them.
        """
        return representation.serializer_repr(
            make_serializer_repr_safe(self),
            indent=1,
        )


class MaybeListSerializer(ReprSafeListSerializer):
    """
    For serializer fields that are allowed to be a reference, an object, or an
    iterable of objects.
    """
    def to_internal_value(self, data):
        if isinstance(data, list):
            return super(MaybeListSerializer, self).to_internal_value(data)
        return self.child.run_validation(data)


class BaseResponseSerializer(serializers.Serializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#responseObject
    """
    description = serializers.CharField()


class HomogenousDictSerializer(ReprSafeSerializer, serializers.Serializer):
    """
    Serializes and object for which all of it's values should be valid against
    a specified serializer class.
    """
    value_serializer_class = None
    value_serializer_kwargs = None
    allow_empty = False

    def __init__(self, *args, **kwargs):
        self.value_serializer_kwargs = kwargs.pop(
            'value_serializer_kwargs',
            self.value_serializer_kwargs or {},
        )
        if self.value_serializer_class is None:
            raise ValueError(
                "Property `value_serializer_class` not declared on {0}".format(
                    type(self),
                ),
            )

        # populate fields for all of the keys in the object to be validated.
        data = kwargs.get('data')
        super(HomogenousDictSerializer, self).__init__(*args, **kwargs)
        self.create_fields_from_data(data)

    def create_fields_from_data(self, data):
        if data:
            fields = [
                key for key, value in data.items() if (value is not None or self.allow_empty)
            ]
            for field_name in fields:
                field_name = str(field_name)
                field = self.value_serializer_class(**self.value_serializer_kwargs)
                self.fields.setdefault(
                    field_name,
                    field,
                )

    def get_value(self, data):
        self.create_fields_from_data(data.get(self.field_name))
        return super(HomogenousDictSerializer, self).get_value(data)


class CommonJSONSchemaSerializer(ReprSafeSerializer, serializers.Serializer):
    default_error_messages = {
        'invalid_type_for_minimum': '`minimum` can only be used for json number types',
        'invalid_type_for_maximum': '`maximum` can only be used for json number types',
        'invalid_type_for_multiple_of': '`multipleOf` can only be used for json number types',
        'invalid_type_for_min_length': '`minLength` can only be used for string types',
        'invalid_type_for_max_length': '`maxLength` can only be used for string types',
        'invalid_type_for_min_items': '`minItems` can only be used for array types',
        'invalid_type_for_max_items': '`maxItems` can only be used for array types',
        'invalid_type_for_unique_items': '`uniqueItems` can only be used for array types',
        'exclusive_minimum_requires_minimum': (
            '`exclusiveMinimum` requires `minimum` to be set'
        ),
        'exclusive_maximum_requires_maximum': (
            '`exclusiveMaximum` requires `maximum` to be set'
        ),
        'enum_must_be_of_array_type': 'enum value must be an array',
    }

    multipleOf = serializers.FloatField(
        allow_null=True, required=False, validators=[MinValueValidator(0)],
    )

    maximum = serializers.FloatField(allow_null=True, required=False)
    exclusiveMaximum = serializers.NullBooleanField(required=False)

    minimum = serializers.FloatField(allow_null=True, required=False)
    exclusiveMinimum = serializers.NullBooleanField(required=False)

    maxLength = serializers.IntegerField(
        allow_null=True, required=False, validators=[MinValueValidator(0)],
    )
    minLength = serializers.IntegerField(
        allow_null=True, required=False, validators=[MinValueValidator(0)],
    )

    pattern = serializers.CharField(allow_null=True, required=False, validators=[regex_validator])

    maxItems = serializers.IntegerField(allow_null=True, required=False)
    minItems = serializers.IntegerField(allow_null=True, required=False)
    uniqueItems = serializers.NullBooleanField(required=False)

    enum = DefaultField(required=False, validators=[is_array_validator])

    class Meta:
        list_serializer_class = ReprSafeListSerializer

    def check_type_for_attr(self, attrs, field_name, types, errors, error_key):
        """
        Shortcut for common pattern of having a keyword that depends on the
        type of the object.  If the types provided do not have any intersection
        with the required types, then an error is created.
        """
        if field_name in attrs and 'type' in attrs:
            declared_types = attrs['type']
            if isinstance(declared_types, six.string_types):
                declared_types = [declared_types]

            if isinstance(types, six.string_types):
                types = [types]

            if not set(types).intersection(declared_types):
                errors[field_name].add_error(
                    self.error_messages[error_key],
                )

    def validate(self, attrs):
        errors = ErrorDict()

        # Minimum
        self.check_type_for_attr(
            attrs,
            'minimum',
            (INTEGER, NUMBER),
            errors,
            'invalid_type_for_minimum',
        )

        if 'exclusiveMinimum' in attrs and 'minimum' not in attrs:
            errors['exclusiveMinimum'].add_error(
                self.error_messages['exclusive_minimum_requires_minimum'],
            )

        # Maximum
        self.check_type_for_attr(
            attrs,
            'maximum',
            (INTEGER, NUMBER),
            errors,
            'invalid_type_for_maximum',
        )

        if 'exclusiveMaximum' in attrs and 'maximum' not in attrs:
            errors['exclusiveMaximum'].add_error(
                self.error_messages['exclusive_maximum_requires_maximum'],
            )

        # multipleOf
        self.check_type_for_attr(
            attrs,
            'multipleOf',
            (INTEGER, NUMBER),
            errors,
            'invalid_type_for_multiple_of',
        )

        # minLength
        self.check_type_for_attr(
            attrs,
            'minLength',
            STRING,
            errors,
            'invalid_type_for_min_length',
        )

        # maxLength
        self.check_type_for_attr(
            attrs,
            'maxLength',
            STRING,
            errors,
            'invalid_type_for_max_length',
        )

        # minItems
        self.check_type_for_attr(
            attrs,
            'minItems',
            ARRAY,
            errors,
            'invalid_type_for_min_items',
        )

        # maxItems
        self.check_type_for_attr(
            attrs,
            'maxItems',
            ARRAY,
            errors,
            'invalid_type_for_max_items',
        )

        # uniqueItems
        self.check_type_for_attr(
            attrs,
            'uniqueItems',
            ARRAY,
            errors,
            'invalid_type_for_unique_items',
        )

        # enum null value special case.
        if 'enum' in attrs and attrs['enum'] is None:
            errors['enum'].add_error(
                self.error_messages['enum_must_be_of_array_type'],
            )

        if errors:
            raise ValidationError(errors)
        return super(CommonJSONSchemaSerializer, self).validate(attrs)


class BaseSchemaSerializer(CommonJSONSchemaSerializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#schemaObject
    """
    _is_unsafe_to_repr = True

    default_error_messages = {
        'invalid_type_for_min_properties': 'minProperties can only be used for `object` types',
        'invalid_type_for_max_properties': 'maxProperties can only be used for `object` types',
    }

    format = serializers.CharField(validators=[format_validator], allow_null=True, required=False)
    title = serializers.CharField(allow_null=True, required=False)
    default = DefaultField(allow_null=True, required=False)

    minProperties = serializers.IntegerField(
        allow_null=True, required=False, validators=[MinValueValidator(0)]
    )
    maxProperties = serializers.IntegerField(
        allow_null=True, required=False, validators=[MinValueValidator(0)],
    )

    required = serializers.NullBooleanField(required=False)
    type = MaybeListCharField(allow_null=True, required=False, validators=[type_validator])

    readOnly = serializers.NullBooleanField(required=False)
    externalDocs = serializers.CharField(allow_null=True, required=False)
    # TODO: how do we do example
    # example =

    # Not Implemented
    # xml
    # discriminator

    def validate(self, attrs):
        errors = ErrorDict()

        # minProperties
        self.check_type_for_attr(
            attrs,
            'minProperties',
            OBJECT,
            errors,
            'invalid_type_for_min_properties',
        )

        # maxProperties
        self.check_type_for_attr(
            attrs,
            'maxProperties',
            OBJECT,
            errors,
            'invalid_type_for_max_properties',
        )

        if errors:
            raise ValidationError(errors)
        return super(BaseSchemaSerializer, self).validate(attrs)

BaseSchemaSerializer._declared_fields['$ref'] = serializers.CharField(
    allow_null=True, required=False,
)


class BaseItemsSerializer(BaseSchemaSerializer):
    def run_validation(self, data):
        if not is_value_of_any_type(data, (ARRAY, OBJECT, STRING)):
            raise ValidationError([[{'non_field_errors': MESSAGES['items']['invalid_type']}]])
        return super(BaseItemsSerializer, self).run_validation(data)

    class Meta:
        list_serializer_class = MaybeListSerializer


class BaseParameterSerializer(TypedDefaultMixin, CommonJSONSchemaSerializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#parameterObject
    """
    default_error_messages = {
        'path_parameters_are_required': (
            "A Parameter who's `in` value is 'path' must be declared as required."
        ),
        'schema_required': (
            "A Parameter who's `in` value is 'body' must declare a schema."
        ),
        'type_required': (
            "A Parameter who's `in` value is not 'body' must declare a type."
        ),
        'collection_format_must_be_multi': (
            "The collectionFormat 'multi' is only valid for `in` values of "
            "\"query\" or \"formData\"."
        ),
        'items_required': (
            "For type \"array\", the items parameter is required."
        ),
    }

    name = serializers.CharField()
    description = serializers.CharField(allow_null=True, required=False)
    required = serializers.NullBooleanField(required=False)

    type = MaybeListCharField(allow_null=True, required=False, validators=[type_validator])
    format = serializers.CharField(validators=[format_validator], allow_null=True, required=False)
    collectionFormat = serializers.CharField(
        allow_null=True, required=False, validators=[collection_format_validator], default=CSV,
    )
    default = DefaultField(allow_null=True, required=False)

    def validate(self, attrs):
        errors = ErrorDict()

        if attrs['in'] == PATH and not attrs.get('required'):
            errors['required'].add_error(
                self.error_messages['path_parameters_are_required'],
            )
        if attrs['in'] == BODY and not attrs.get('schema'):
            errors['schema'].add_error(
                self.error_messages['schema_required'],
            )
        if attrs['in'] != BODY:
            if 'type' not in attrs:
                errors['type'].add_error(
                    self.error_messages['type_required'],
                )
            if attrs.get('collectionFormat') == MULTI:
                if attrs['in'] not in (QUERY, FORM_DATA):
                    errors['collectionFormat'].add_error(
                        self.error_messages['collection_format_must_be_multi'],
                    )
            self.validate_default_type(attrs, errors)

        if attrs.get('type') == ARRAY and not attrs.get('items'):
            errors['items'].add_error(
                self.error_messages['items_required'],
            )

        if errors:
            raise ValidationError(errors)

        return super(BaseParameterSerializer, self).validate(attrs)

    def create(self, validated_data):
        return validated_data


class BaseHeaderSerializer(TypedDefaultMixin, CommonJSONSchemaSerializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#header-object-
    """
    default_error_messages = {
        'items_required': (
            "When type is \"array\" the \"items\" is required"
        ),
    }
    description = serializers.CharField(allow_null=True, required=False)
    type = serializers.CharField(validators=[header_type_validator])
    format = serializers.CharField(validators=[format_validator], allow_null=True, required=False)
    collectionFormat = serializers.CharField(
        allow_null=True, required=False, validators=[collection_format_validator], default=CSV,
    )
    default = DefaultField(allow_null=True, required=False)

    def validate(self, attrs):
        errors = ErrorDict()

        if attrs.get('type') == ARRAY and 'items' not in attrs:
            errors['items'].add_error(
                self.error_messages['items_required'],
            )
        self.validate_default_type(attrs, errors)

        if errors:
            raise ValidationError(errors)
        return super(BaseHeaderSerializer, self).validate(attrs)

    def create(self, validated_data):
        return validated_data


# Cannot declare this as a property on the class because `in` is a reserved word.
BaseParameterSerializer._declared_fields['in'] = serializers.CharField(
    validators=[parameter_in_validator],
)



import functools
import operator

from flex.exceptions import ValidationError
from flex.utils import chain_reduce_partial
from flex.context_managers import ErrorCollection
from flex.validation.common import (
    validate_object,
    validate_path_to_api_path,
    validate_request_method_to_operation,
)
from flex.validation.schema import construct_schema_validators
from flex.error_messages import MESSAGES
from flex.constants import (
    EMPTY,
    PATH,
)
from flex.parameters import (
    filter_parameters,
    merge_parameter_lists,
    dereference_parameter_list,
)
from flex.validation.header import (
    construct_header_validators,
)
from flex.validation.path import (
    generate_path_parameters_validator,
)
from flex.validation.common import (
    generate_value_processor,
)
from flex.http import Response


def validate_status_code_to_response_definition(response, operation_definition):
    """
    Given a response, validate that the response status code is in the accepted
    status codes defined by this endpoint.

    If so, return the response definition that corresponds to the status code.
    """
    status_code = response.status_code
    operation_responses = operation_definition['responses']
    try:
        response_definition = operation_responses[status_code]
    except KeyError:
        raise ValidationError(
            MESSAGES['response']['invalid_status_code'].format(
                status_code, operation_responses.keys(),
            ),
        )
    return response_definition


def generate_response_body_validator(schema, context, **kwargs):
    validators = construct_schema_validators(schema, context=context)
    return chain_reduce_partial(
        operator.attrgetter('data'),
        functools.partial(
            validate_object,
            validators=validators,
        ),
    )


def generate_response_header_validator(headers, context, **kwargs):
    validators = {}
    for key, header_definition in headers.items():
        # generate a function that will attempt to cast the header to the
        # appropriate type.
        header_processor = generate_value_processor(
            context=context,
            **header_definition
        )
        # generate a function that will validate the header.
        header_validator = functools.partial(
            validate_object,
            validators=construct_header_validators(header_definition, context=context),
        )
        # Chain the type casting function, the individual header validation
        # function with a methodcaller that will fetch the header with
        # `response.headers.get(header_name, EMPTY)` and then feed that into
        # the type casting function and then into the validation function.
        validators[key] = chain_reduce_partial(
            operator.methodcaller('get', key, EMPTY),
            header_processor,
            header_validator,
        )
    return chain_reduce_partial(
        operator.attrgetter('headers'),
        functools.partial(validate_object, validators=validators),
    )


def validate_response_content_type(response, content_types):
    assert isinstance(response, Response)  # TODO: remove this sanity check
    if response.content_type and response.content_type not in content_types:
        raise ValidationError(
            MESSAGES['content_type']['invalid'].format(
                response.content_type, content_types,
            ),
        )


def generate_response_content_type_validator(produces, **kwargs):
    return functools.partial(
        validate_response_content_type,
        content_types=produces,
    )


def generate_path_validator(api_path, path_definition, parameters,
                            context, **kwargs):
    """
    Generates a callable for validating the parameters in a response object.
    """
    global_parameters = context.get('parameters', {})
    path_level_parameters = dereference_parameter_list(
        path_definition.get('parameters', []),
        global_parameters,
    )
    operation_level_parameters = dereference_parameter_list(
        parameters,
        global_parameters,
    )

    all_parameters = merge_parameter_lists(
        path_level_parameters,
        operation_level_parameters,
    )

    # PATH
    in_path_parameters = filter_parameters(all_parameters, in_=PATH)
    return chain_reduce_partial(
        operator.attrgetter('path'),
        generate_path_parameters_validator(api_path, in_path_parameters, context),
    )


validator_mapping = {
    # TODO: tests for parameters AND figure out how to do parameters.
    'headers': generate_response_header_validator,
    'produces': generate_response_content_type_validator,
    'schema': generate_response_body_validator,
}


def generate_response_validator(api_path, operation_definition, response_definition,
                                path_definition, context):
    validators = {}

    # Parameters is special cause it needs data from both the
    # `operation_definition` and the `path_definition`
    validators['path'] = generate_path_validator(
        api_path=api_path,
        path_definition=path_definition,
        parameters=operation_definition.get('parameters', []),
        context=context,
    )

    for key in validator_mapping:
        if key in response_definition:
            validators[key] = validator_mapping[key](context=context, **response_definition)
        elif key in operation_definition:
            validators[key] = validator_mapping[key](context=context, **operation_definition)
        elif key in path_definition:
            validators[key] = validator_mapping[key](context=context, **path_definition)

    if 'produces' in context and 'produces' not in validators:
        validators['produces'] = generate_response_content_type_validator(**context)

    return functools.partial(
        validate_object,
        validators=validators,
    )


def validate_response(response, request_method, schema):
    """
    Response validation involves the following steps.
       4. validate that the response status_code is in the allowed responses for
          the request method.
       5. validate that the response content validates against any provided
          schemas for the responses.
       6. headers, content-types, etc..., ???
    """
    with ErrorCollection() as errors:
        # 1
        # TODO: tests
        try:
            api_path = validate_path_to_api_path(
                path=response.path,
                **schema
            )
        except ValidationError as err:
            errors['path'].extend(list(err.messages))
            return  # this causes an exception to be raised since errors is no longer falsy.

        path_definition = schema['paths'][api_path] or {}

        # TODO: tests
        try:
            operation_definition = validate_request_method_to_operation(
                request_method=request_method,
                path_definition=path_definition,
            )
        except ValidationError as err:
            errors['method'].append(err.message)
            return

        # 4
        try:
            response_definition = validate_status_code_to_response_definition(
                response=response,
                operation_definition=operation_definition,
            )
        except ValidationError as err:
            errors['status_code'].add_error(err.detail)
        else:
            # 5
            response_validator = generate_response_validator(
                api_path,
                operation_definition=operation_definition,
                path_definition=path_definition,
                response_definition=response_definition,
                context=schema,
            )
            try:
                response_validator(response)
            except ValidationError as err:
                errors['body'].add_error(err.detail)



import rest_framework

__all__ = [
    'HomogenousDictSerializer',
    'BaseResponseSerializer',
    'BaseParameterSerializer',
    'BaseSchemaSerializer',
    'BaseItemsSerializer',
    'BaseHeaderSerializer',
]

if rest_framework.__version__ >= '3.0.0':
    from .drf_3.common import (
        HomogenousDictSerializer,
        BaseResponseSerializer,
        BaseParameterSerializer,
        BaseSchemaSerializer,
        BaseItemsSerializer,
        BaseHeaderSerializer,
    )
else:
    from .drf_2.common import (
        BaseHeaderSerializer,
        BaseItemsSerializer,
        BaseParameterSerializer,
        BaseResponseSerializer,
        BaseSchemaSerializer,
        HomogenousDictSerializer,
    )



import rest_framework

__all__ = [
    'SecurityRequirementReferenceField',
    'MaybeListCharField',
]

if rest_framework.__version__ >= '3.0.0':
    from .drf_3.fields import (
        SecurityRequirementReferenceField,
        MaybeListCharField,
    )
else:
    from .drf_2.fields import (
        SecurityRequirementReferenceField,
        MaybeListCharField,
    )



import rest_framework

__all__ = [
    'SwaggerDefinitionsSerializer',
    'DefinitionsSerializer',
    'SecuritySchemeSerializer',
    'SchemaSerializer',
    'HeaderSerializer',
    'ParameterDefinitionsSerializer',
    'ItemsSerializer',
]

if rest_framework.__version__ >= '3.0.0':
    from .drf_3.definitions import (
        SwaggerDefinitionsSerializer,
        DefinitionsSerializer,
        SecuritySchemeSerializer,
        SchemaSerializer,
        HeaderSerializer,
        ParameterDefinitionsSerializer,
        ItemsSerializer,
    )
else:
    from .drf_2.definitions import (
        SwaggerDefinitionsSerializer,
        DefinitionsSerializer,
        SecuritySchemeSerializer,
        SchemaSerializer,
        HeaderSerializer,
        ParameterDefinitionsSerializer,
        ItemsSerializer,
    )



import functools

from flex.exceptions import ValidationError
from flex.utils import (
    is_non_string_iterable,
    is_value_of_any_type,
)
from flex.constants import EMPTY

from django.core.exceptions import ValidationError as DjangoValidationError


def partial_safe_wraps(wrapped_func, *args, **kwargs):
    """
    A version of `functools.wraps` that is safe to wrap a partial in.
    """
    if isinstance(wrapped_func, functools.partial):
        return partial_safe_wraps(wrapped_func.func)
    else:
        return functools.wraps(wrapped_func)


def maybe_iterable(func):
    @partial_safe_wraps(func)
    def inner(value):
        if is_non_string_iterable(value):
            return list(map(func, value))
        else:
            return func(value)
    return inner


def skip_if_not_of_type(*types):
    def outer(func):
        @partial_safe_wraps(func)
        def inner(value, *args, **kwargs):
            if value is EMPTY or is_value_of_any_type(value, types):
                return func(value, *args, **kwargs)
        return inner
    return outer


def skip_if_empty(func):
    """
    Decorator for validation functions which makes them pass if the value
    passed in is the EMPTY sentinal value.
    """
    @partial_safe_wraps(func)
    def inner(value, *args, **kwargs):
        if value is EMPTY:
            return
        else:
            return func(value, *args, **kwargs)
    return inner


RESERVED_WORDS = (
    'in',
    'format',
    'type',
)


def rewrite_reserved_words(func):
    """
    Given a function whos kwargs need to contain a reserved word such as 'in',
    allow calling that function with the keyword as 'in_', such that function
    kwargs are rewritten to use the reserved word.
    """
    @partial_safe_wraps(func)
    def inner(*args, **kwargs):
        for word in RESERVED_WORDS:
            key = "{0}_".format(word)
            if key in kwargs:
                kwargs[word] = kwargs.pop(key)
        return func(*args, **kwargs)
    return inner


def suffix_reserved_words(func):
    """
    Given a function that is called with a reseved word, rewrite the keyword
    with an underscore suffix.
    """
    @partial_safe_wraps(func)
    def inner(*args, **kwargs):
        for word in RESERVED_WORDS:
            if word in kwargs:
                key = "{0}_".format(word)
                kwargs[key] = kwargs.pop(word)
        return func(*args, **kwargs)
    return inner


def translate_validation_error(func):
    """
    Given a function that potentially raises
    `django.core.exceptions.ValidationError`, reraise the same error as a
    `flex.exceptions.ValidationError`.
    """
    @partial_safe_wraps(func)
    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except DjangoValidationError as err:
            if isinstance(err, ValidationError):
                raise
            raise ValidationError(err.messages)
    return inner



import six

from rest_framework import serializers

from flex.exceptions import (
    ValidationError,
    ErrorDict,
)
from flex.serializers.validators import (
    security_type_validator,
    security_api_key_location_validator,
    security_flow_validator,
)
from flex.constants import (
    API_KEY,
    OAUTH_2,
    IMPLICIT,
    PASSWORD,
    APPLICATION,
    ACCESS_CODE,
)

from .common import (
    HomogenousDictSerializer,
    BaseResponseSerializer,
    BaseSchemaSerializer,
    BaseParameterSerializer,
    BaseItemsSerializer,
    BaseHeaderSerializer,
)


class SchemaSerializer(BaseSchemaSerializer):
    def to_internal_value(self, data):
        if isinstance(data, six.string_types):
            self.context['deferred_references'].add(data)
            return data
        elif '$ref' in data:
            self.context['deferred_references'].add(data['$ref'])
        return super(SchemaSerializer, self).to_internal_value(data)


class DefinitionsSerializer(HomogenousDictSerializer):
    default_error_messages = {
        'unknown_references': "Unknown references `{0}`",
    }

    value_serializer_class = SchemaSerializer

    def validate(self, attrs):
        deferred_references = self.context.get('deferred_references', set())
        missing_references = deferred_references.difference(attrs.keys())
        if missing_references:
            raise ValidationError(
                self.error_messages['unknown_references'].format(
                    list(missing_references),
                ),
            )
        return super(DefinitionsSerializer, self).validate(attrs)

    def create(self, validated_data):
        return validated_data


class PropertiesSerializer(HomogenousDictSerializer):
    value_serializer_class = SchemaSerializer


class ItemsSerializer(BaseItemsSerializer):
    def run_validation(self, data):
        if isinstance(data, six.string_types):
            value = self.to_internal_value(data)
            return value
        return super(ItemsSerializer, self).run_validation(data)

    def to_internal_value(self, data):
        if isinstance(data, six.string_types):
            self.context['deferred_references'].add(data)
            return data
        return super(ItemsSerializer, self).to_internal_value(data)


# These fields include recursive use of the `SchemaSerializer` so they have to
# be attached after the `SchemaSerializer` class has been created.
SchemaSerializer._declared_fields['properties'] = PropertiesSerializer(required=False)
SchemaSerializer._declared_fields['items'] = ItemsSerializer(required=False, many=True)
SchemaSerializer._declared_fields['allOf'] = SchemaSerializer(required=False, many=True)


class HeaderSerializer(BaseHeaderSerializer):
    items = ItemsSerializer(required=False, many=True)
    schema = SchemaSerializer(required=False)


class HeadersSerializer(HomogenousDictSerializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#headersObject
    """
    value_serializer_class = HeaderSerializer


class ParameterSerializer(BaseParameterSerializer):
    schema = SchemaSerializer(required=False)
    items = ItemsSerializer(required=False, many=True)


class ParameterDefinitionsSerializer(HomogenousDictSerializer):
    value_serializer_class = ParameterSerializer


class ScopesSerializer(HomogenousDictSerializer):
    value_serializer_class = serializers.CharField


class SecuritySchemeSerializer(serializers.Serializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#securityDefinitionsObject
    """
    default_error_messages = {
        'name_required': 'When type is "apiKey", "name" is required.',
        'in_required': 'When type is "apiKey", "in" is required.',
        'flow_required': 'When type is "oath2" flow is required.',
        'authorization_url_required': (
            'When type is "oath2" and flow is one of ("implicit", '
            '"accessCode"), "authorizationUrl" is required.'
        ),
        'token_url_required': (
            'When type is "oath2" and flow is one of ("password", '
            '"application", "accessCode"), "authorizationUrl" is required.'
        ),
    }
    type = serializers.CharField(validators=[security_type_validator])
    description = serializers.CharField(required=False)
    name = serializers.CharField(required=False)
    flow = serializers.CharField(
        required=False, validators=[security_flow_validator],
    )
    # TODO: support SHOULD clauses about these being in the form of a url.
    authorizationUrl = serializers.CharField(required=False)
    tokenUrl = serializers.CharField(required=False)

    scopes = ScopesSerializer(required=False)

    def validate(self, attrs):
        errors = ErrorDict()

        # apiKey validations
        if attrs['type'] == API_KEY:
            if 'name' not in attrs:
                errors['name'].add_error(
                    self.error_messages['name_required'],
                )
            if 'in' not in attrs:
                errors['in'].add_error(
                    self.error_messages['in_required'],
                )
        elif attrs['type'] == OAUTH_2:
            if 'flow' not in attrs:
                errors['flow'].add_error(
                    self.error_messages['flow_required'],
                )
            if attrs.get('flow') in (IMPLICIT, ACCESS_CODE):
                if 'authorizationUrl' not in attrs:
                    errors['authorizationUrl'].add_error(
                        self.error_messages['authorization_url_required'],
                    )
            if attrs.get('flow') in (PASSWORD, APPLICATION, ACCESS_CODE):
                if 'tokenUrl' not in attrs:
                    errors['tokenUrl'].add_error(
                        self.error_messages['token_url_required'],
                    )

        if errors:
            raise ValidationError(errors)
        return super(SecuritySchemeSerializer, self).validate(attrs)


SecuritySchemeSerializer._declared_fields['in'] = serializers.CharField(
    required=False,
    validators=[security_api_key_location_validator],
)


class SecurityDefinitionsSerializer(HomogenousDictSerializer):
    value_serializer_class = SecuritySchemeSerializer


class ResponseSerializer(BaseResponseSerializer):
    """
    https://github.com/wordnik/swagger-spec/blob/master/versions/2.0.md#responseObject
    """
    schema = SchemaSerializer(required=False)
    headers = HeadersSerializer(required=False)
    # example  # TODO: how do we do example.


class ResponseDefinitionsSerializer(HomogenousDictSerializer):
    value_serializer_class = ResponseSerializer


class SwaggerDefinitionsSerializer(serializers.Serializer):
    """
    Step 1 in the schema validation process is to gather all of the
    definitions.
    """
    def __init__(self, *args, **kwargs):
        context = kwargs.pop('context', {})
        context['deferred_references'] = set()
        kwargs['context'] = context
        super(SwaggerDefinitionsSerializer, self).__init__(*args, **kwargs)

    definitions = DefinitionsSerializer(required=False)
    parameters = ParameterDefinitionsSerializer(required=False)
    securityDefinitions = SecurityDefinitionsSerializer(required=False)
    responses = ResponseDefinitionsSerializer(required=False)

    def validate(self, attrs):
        deferred_references = self.context['deferred_references']
        missing_references = set()
        for deferred_reference in deferred_references:
            for definitions in attrs.values():
                if deferred_reference in definitions:
                    break
            else:
                missing_references.add(deferred_reference)
        if missing_references:
            raise ValidationError(
                {'missing_references': list(missing_references)},
            )
        return super(SwaggerDefinitionsSerializer, self).validate(attrs)

    def create(self, validated_data):
        return validated_data

    def update(self, instance, validated_data):
        instance.update(validated_data)
        return instance



from __future__ import unicode_literals

import six
import collections

from flex.utils import (
    is_non_string_iterable,
    prettify_errors,
)

from rest_framework.serializers import ValidationError as DRFValidationError


class ErrorList(list):
    def __init__(self, value=None):
        super(ErrorList, self).__init__()
        if value:
            self.add_error(value)

    def add_error(self, error):
        if is_non_string_iterable(error) and not isinstance(error, collections.Mapping):
            for value in error:
                self.add_error(value)
        else:
            self.append(error)


class ErrorDict(collections.defaultdict):
    def __init__(self, value=None):
        super(ErrorDict, self).__init__(ErrorList)
        for k, v in (value or {}).items():
            self.add_error(k, v)

    def add_error(self, key, error):
        self[key].add_error(error)


class ValidationError(DRFValidationError):
    def __init__(self, error):
        if not isinstance(error, collections.Mapping) and \
           is_non_string_iterable(error) and \
           len(error) == 1:
            error = error[0]
        self._error = error

        if isinstance(self._error, collections.Mapping):
            self.error_dict = self._error
        elif is_non_string_iterable(self._error):
            self.error_list = self._error

    def __repr__(self):
        return "<type '{0}'>, {1}".format(
            repr(type(self)),
            repr(self.detail),
        )

    def __str__(self):
        return prettify_errors(self._error)

    @property
    def detail(self):
        if isinstance(self._error, six.string_types):
            return [self._error]
        elif isinstance(self._error, collections.Mapping):
            return self._error
        return self._error

    @property
    def messages(self):
        if isinstance(self._error, six.string_types):
            return [self._error]
        elif isinstance(self._error, collections.Mapping):
            return [self._error]
        return self._error



import collections
import functools
import re

from flex.constants import (
    PATH,
)
from flex.parameters import (
    find_parameter,
    merge_parameter_lists,
    dereference_parameter_list,
)


REGEX_REPLACEMENTS = (
    ('\.', '\.'),
    ('\{', '\{'),
    ('\}', '\}'),
)


def escape_regex_special_chars(api_path):
    """
    Turns the non prametrized path components into strings subtable for using
    as a regex pattern.  This primarily involves escaping special characters so
    that the actual character is matched in the regex.
    """
    def substitute(string, replacements):
        pattern, repl = replacements
        return re.sub(pattern, repl, string)

    return functools.reduce(substitute, REGEX_REPLACEMENTS, api_path)


# matches the parametrized parts of a path.
# eg. /{id}/ matches the `{id}` part of it.
PARAMETER_REGEX = re.compile('(\{[^\}]+})')


def construct_parameter_pattern(parameter):
    """
    Given a parameter definition returns a regex pattern that will match that
    part of the path.
    """
    name = parameter['name']

    return "(?P<{name}>.+)".format(name=name)


def process_path_part(part, parameters):
    """
    Given a part of a path either:
        - If it is a parameter:
            parse it to a regex group
        - Otherwise:
            escape any special regex characters
    """
    if PARAMETER_REGEX.match(part):
        parameter_name = part.strip('{}')
        try:
            parameter = find_parameter(parameters, name=parameter_name, in_=PATH)
        except ValueError:
            pass
        else:
            return construct_parameter_pattern(parameter)
    return escape_regex_special_chars(part)


def get_parameter_names_from_path(api_path):
    return tuple(p.strip('{}') for p in PARAMETER_REGEX.findall(api_path))


def path_to_pattern(api_path, parameters):
    """
    Given an api path, possibly with parameter notation, return a pattern
    suitable for turing into a regular expression which will match request
    paths that conform to the parameter definitions and the api path.
    """
    parts = re.split(PARAMETER_REGEX, api_path)
    pattern = ''.join((process_path_part(part, parameters) for part in parts))

    if not pattern.startswith('^'):
        pattern = "^{0}".format(pattern)
    if not pattern.endswith('$'):
        pattern = "{0}$".format(pattern)

    return pattern


def path_to_regex(api_path, path_parameters, global_parameters=None):
    if global_parameters is None:
        global_parameters = {}
    pattern = path_to_pattern(
        api_path=api_path,
        parameters=merge_parameter_lists(
            global_parameters.values(),
            dereference_parameter_list(path_parameters, global_parameters),
        ),
    )
    return re.compile(pattern)


def match_path_to_api_path(path_definitions, target_path, base_path='', global_parameters=None):
    """
    Match a request or response path to one of the api paths.

    Anything other than exactly one match is an error condition.
    """
    if global_parameters is None:
        global_parameters = {}
    assert isinstance(global_parameters, collections.Mapping)
    if target_path.startswith(base_path):
        target_path = target_path[len(base_path):]

    # Convert all of the api paths into Path instances for easier regex matching.
    paths = {
        p: path_to_regex(
            api_path=p,
            path_parameters=(v or {}).get('parameters', []),
            global_parameters=global_parameters,
        )
        for p, v in path_definitions.items()
    }

    matches = [p for p, r in paths.items() if r.match(target_path)]

    if not matches:
        raise LookupError('No paths found for {0}'.format(target_path))
    elif len(matches) > 1:
        raise LookupError('Multipue paths found for {0}.  Found `{1}`'.format(
            target_path, matches,
        ))
    else:
        return matches[0]



from rest_framework import serializers

from flex.exceptions import ValidationError
from flex.error_messages import MESSAGES
from flex.utils import is_non_string_iterable

from .mixins import TranslateValidationErrorMixin


class MaybeListCharField(TranslateValidationErrorMixin, serializers.CharField):
    def from_native(self, value):
        if is_non_string_iterable(value):
            return value
        return super(MaybeListCharField, self).from_native(value)


class SecurityRequirementReferenceField(serializers.CharField):
    """
    Field that references a defined security scheme declared in the Security
    Definitions.
    """
    default_error_messages = {
        'unknown_reference': "Unknown Security Scheme reference `{0}`",
    }

    def validate(self, value):
        if value not in self.context.get('securityDefinitions', {}):
            raise ValidationError(
                MESSAGES['unknown_reference']['security'].format(value),
            )



from __future__ import unicode_literals


TYPE_MESSAGES = {
    'unknown': 'Unknown type: {0}',
    'invalid': "Got value `{0}` of type `{1}`.  Value must be of type(s): `{2}`",
    'invalid_header_type': (
        "Invalid type for header: `{0}`.  Must be one of 'string', 'number', "
        "'integer', 'boolean', or 'array'."
    ),
}

FORMAT_MESSAGES = {
    'invalid': "Value {0} does not conform to the format {1}",
    'invalid_uuid': "{0} is not a valid uuid",
    'invalid_datetime': "{0} is not a valid iso8601 date-time",
    'too_many_bits': "Integer {0} has {1} bits.  Must be no more than {2} bits",
    'invalid_uri': "The value `{0}` is not valid according to RFC3987.",
    'invalid_email': "The email address `{0}` is invalid according to RFC5322.",
}

REQUIRED_MESSAGES = {
    'required': "This value is required",
}

MULTIPLE_OF_MESSAGES = {
    'invalid': "Value must be a multiple of {0}.  Got {1} which is not.",
}

MINIMUM_AND_MAXIMUM_MESSAGES = {
    'invalid': "{0} must be {1} than {2}",
}


MIN_ITEMS_MESSAGES = {
    'invalid': "Array must have at least {0} items.  It had only had {1} items.",
}


MAX_ITEMS_MESSAGES = {
    'invalid': "Array must have no more than {0} items.  It had {1} items.",
}


UNIQUE_ITEMS_MESSAGES = {
    'invalid': "Array items must be unique.  The following items appeard more than once: {0}",
}


ENUM_MESSAGES = {
    'invalid': "Invalid value.  {0} is not one of the available options ({1})",
}


PATTERN_MESSAGES = {
    'invalid': "{0} did not match the pattern `{1}`.",
}


MIN_PROPERTIES_MESSAGES = {
    'invalid': "Object must have more than {0} properties.  It had {1}",
}


MAX_PROPERTIES_MESSAGES = {
    'invalid': "Object must have less than {0} properties.  It had {1}",
}


ITEMS_MESSAGES = {
    'invalid_type': '`items` must be a reference, a schema, or an array of schemas.',
}


REQUEST_MESSAGES = {
    'invalid_method': (
        'Request was not one of the allowed request methods.  Got '
        '`{0}`: Expected one of: `{1}`'
    ),
}


RESPONSE_MESSAGES = {
    'invalid_status_code': (
        "Request status code was not found in the known response codes.  Got "
        "`{0}`: Expected one of: `{1}`"
    )
}


PATH_MESSAGES = {
    'unknown_path': 'Request path did not match any of the known api paths.',
    'missing_parameter': (
        "The parameter named `{0}` is declared to be a PATH parameter but does "
        "not appear in the api path `{1}`.  All path parameters must exist as a "
        "parameter in the api path"
    )
}


UNKNOWN_REFERENCE_MESSAGES = {
    'security': "Unknown SecurityScheme reference `{0}`",
    'parameter': "Unknown Parameter reference `{0}`",
    'definition': 'Unknown definition reference `{0}`',
}


CONTENT_TYPE_MESSAGES = {
    'invalid': 'Invalid content type `{0}`.  Must be one of `{1}`.',
}


MESSAGES = {
    'type': TYPE_MESSAGES,
    'format': FORMAT_MESSAGES,
    'required': REQUIRED_MESSAGES,
    'multiple_of': MULTIPLE_OF_MESSAGES,
    'minimum': MINIMUM_AND_MAXIMUM_MESSAGES,
    'maximum': MINIMUM_AND_MAXIMUM_MESSAGES,
    'min_items': MIN_ITEMS_MESSAGES,
    'max_items': MAX_ITEMS_MESSAGES,
    'min_properties': MIN_PROPERTIES_MESSAGES,
    'max_properties': MAX_PROPERTIES_MESSAGES,
    'unique_items': UNIQUE_ITEMS_MESSAGES,
    'enum': ENUM_MESSAGES,
    'pattern': PATTERN_MESSAGES,
    'items': ITEMS_MESSAGES,
    'request': REQUEST_MESSAGES,
    'response': RESPONSE_MESSAGES,
    'path': PATH_MESSAGES,
    'unknown_reference': UNKNOWN_REFERENCE_MESSAGES,
    'content_type': CONTENT_TYPE_MESSAGES,
}



import collections

from flex.decorators import rewrite_reserved_words


@rewrite_reserved_words
def is_match(parameter, **kwargs):
    for key, value in kwargs.items():
        if key not in parameter:
            return False
        elif parameter[key] != value:
            return False
    return True


@rewrite_reserved_words
def filter_parameters(parameters, **kwargs):
    return [p for p in parameters if is_match(p, **kwargs)]


@rewrite_reserved_words
def find_parameter(parameters, **kwargs):
    """
    Given a list of parameters, find the one with the given name.
    """
    matching_parameters = filter_parameters(parameters, **kwargs)
    if len(matching_parameters) == 1:
        return matching_parameters[0]
    elif len(matching_parameters) > 1:
        raise ValueError("More than 1 parameter matched")
    raise ValueError("No parameters matched")


def merge_parameter_lists(*parameter_definitions):
    """
    Merge multiple lists of parameters into a single list.  If there are any
    duplicate definitions, the last write wins.
    """
    merged_parameters = {}
    for parameter_list in parameter_definitions:
        for parameter in parameter_list:
            key = (parameter['name'], parameter['in'])
            merged_parameters[key] = parameter
    return merged_parameters.values()


def dereference_parameter_list(parameters, parameter_definitions):
    return [
        (p if isinstance(p, collections.Mapping) else parameter_definitions[p])
        for p in parameters
    ]



from flex.utils import is_value_of_type
from flex.decorators import (
    translate_validation_error,
)


class TypedDefaultMixin(object):
    default_error_messages = {
        'default_is_incorrect_type': (
            "The value supplied for 'default' must match the specified type."
        ),
    }

    def validate_default_type(self, attrs, errors):
        if 'default' in attrs and 'type' in attrs:
            if not is_value_of_type(attrs['default'], attrs['type']):
                errors['default'].add_error(
                    self.error_messages['default_is_incorrect_type'],
                )


class TranslateValidationErrorMixin(object):
    @translate_validation_error
    def field_from_native(self, *args, **kwargs):
        return super(TranslateValidationErrorMixin, self).field_from_native(
            *args, **kwargs
        )

    @translate_validation_error
    def validate(self, *args, **kwargs):
        return super(TranslateValidationErrorMixin, self).validate(
            *args, **kwargs
        )
